VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cCamera"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False

Option Explicit

' a user friendly name for the camera
Public Tag As String
' the position of the camera
Private Orig As D3DVECTOR
' the direction it is facing
Private Orient As D3DVECTOR
' the velocity of the camera
Private Vel As D3DVECTOR
' the spin of the camera
Private Spn As D3DVECTOR
' the field of view
Public FOV As Single
' clipping distances
Public NearClip As Single
Public FarClip As Single ' (like redraw distance, but this also affects perspective transformations)
' aspect ratio (height/width of projected view)
Public Aspect As Single

' the frame the camera is attached to, if any
Public Frame As cFrame

' the mode the camera can be in
Enum CAMERA_MODE
    CM_DEFAULT ' use actual coords and rotation in camera
    CM_LOOK ' look at the orient vector
    CM_LOOKATFRAME ' use actual position, but automatically look at a frame
    CM_WELDTOFRAME ' attaches a camera to a frame, so that it follows it automatically
End Enum

' the mode the camera is in
Public Mode As CAMERA_MODE








' set some defaults
Private Sub Class_Initialize()
On Error Resume Next
    FOV = PI / 3
    NearClip = 0.1
    FarClip = 10000
    Aspect = 0.75
    Mode = CM_DEFAULT
End Sub



' the origin
Property Get Origin() As D3DVECTOR
    Origin = Orig
End Property
Property Let Origin(vec As D3DVECTOR)
    Orig = vec
End Property

' the orientation
Property Get Orientation() As D3DVECTOR
    Orientation = Orient
End Property
Property Let Orientation(vec As D3DVECTOR)
    Orient = vec
End Property

' the velocity
Property Get Velocity() As D3DVECTOR
    Velocity = Vel
End Property
Property Let Velocity(vec As D3DVECTOR)
    Vel = vec
End Property

' the spin
Property Get Spin() As D3DVECTOR
    Spin = Spn
End Property
Property Let Spin(vec As D3DVECTOR)
    Spn = vec
End Property



' moves the camera by or to a vector
Sub Move(Translation As D3DVECTOR)
On Error Resume Next
    D3DXVec3Add Orig, Orig, Translation
End Sub
Sub MoveTo(Position As D3DVECTOR)
On Error Resume Next
    Orig = Position
End Sub



' gets the translation that would happen if the camera were to
' move in the direction it faces
Function GetFowardVector(Distance As Single, Optional timeElapsed As Single = 1) As D3DVECTOR
Dim matView As D3DMATRIX, matTrans As D3DMATRIX, matRot As D3DMATRIX, matRotX As D3DMATRIX, matRotY As D3DMATRIX
Dim Length As Single
Dim Translation As D3DVECTOR
On Error Resume Next
    ' rotate x, y and z
    D3DXMatrixRotationY matRotY, -Orient.y
    D3DXMatrixRotationX matRotX, -Orient.x
    ' multiply matrices together to create rotation matrix
    D3DXMatrixMultiply matRot, matRotX, matRotY
    ' translate
    D3DXMatrixTranslation matTrans, 0, 0, Distance * timeElapsed
    ' multiply that translation and rotation
    D3DXMatrixMultiply matView, matTrans, matRot
    ' translate by matrix
    D3DXVec3TransformCoord Translation, Translation, matView
    ' return translation
    GetFowardVector = Translation
End Function



' moves the camera fowards
Sub MoveFoward(Distance As Single, Optional timeElapsed As Single = 0)
Dim Translation As D3DVECTOR
On Error Resume Next
    Mode = CM_DEFAULT
    Move GetFowardVector(Distance, timeElapsed)
End Sub



' rotates to vector
Sub RotateTo(vec As D3DVECTOR)
On Error Resume Next
    Mode = CM_DEFAULT
    Orient = vec
End Sub
' rotates by vector
Sub Rotate(vec As D3DVECTOR)
On Error Resume Next
    Mode = CM_DEFAULT
    D3DXVec3Add Orient, Orient, vec
End Sub



' looks at a position
Sub LookAt(vec As D3DVECTOR)
On Error Resume Next
    Mode = CM_LOOK
    Orient = vec
End Sub



' sets the variables which affect the view and projection transformation
Sub SetViewAndProjection(FOV As Single, NearClip As Single, FarClip As Single, Aspect As Single)
On Error Resume Next
    With Me
        .FOV = FOV
        .NearClip = NearClip
        .FarClip = FarClip
        .Aspect = Aspect
    End With
End Sub



' get camera to look at a frame
Sub LookAtFrame(theFrame As cFrame)
On Error Resume Next
    Set Frame = theFrame
    Mode = CM_LOOKATFRAME
End Sub



' attaches camera to a frame
Sub WeldToFrame(theFrame As cFrame)
On Error Resume Next
    Set Frame = theFrame
    Mode = CM_WELDTOFRAME
End Sub



' updates the camera location (and the listener)
Sub RefreshLocation(timeElapsed As Single)
Dim vec As D3DVECTOR
On Error Resume Next
    D3DXVec3Scale vec, Vel, timeElapsed
    D3DXVec3Add Orig, Orig, vec
    D3DXVec3Scale vec, Spn, timeElapsed
    D3DXVec3Add Orient, Orient, vec
    With Listener
        .SetPosition Orig, DS3D_DEFERRED
        .SetOrientation Vector3(0, 1, 0), Orient, DS3D_DEFERRED
        .SetVelocity Vel, DS3D_DEFERRED
        .Update
    End With
End Sub



' sets the camera as the current view and projection transformation
' looks at the orient vector or rotates to the look vector
Sub SetAsCurrent()
Dim matView As D3DMATRIX, matTrans As D3DMATRIX, matRot As D3DMATRIX, matRotX As D3DMATRIX, matRotY As D3DMATRIX, matRotZ As D3DMATRIX
Dim matProjection As D3DMATRIX
Dim vec As D3DVECTOR, Ray As D3DVECTOR
On Error Resume Next
    Select Case Mode
        Case CM_DEFAULT ' use actual coords and rotation
            ' rotate x, y and z
            D3DXMatrixRotationY matRotY, Orient.y
            D3DXMatrixRotationX matRotX, Orient.x
            D3DXMatrixRotationX matRotZ, Orient.z
            ' multiply matrices together to create rotation matrix
            D3DXMatrixMultiply matRot, matRotX, matRotY
            D3DXMatrixMultiply matRot, matRot, matRotZ
            ' translate
            D3DXMatrixTranslation matTrans, -Orig.x, -Orig.y, -Orig.z
            ' multiply that translation and rotation
            D3DXMatrixMultiply matView, matTrans, matRot
        Case CM_LOOK ' look at orient vector
            D3DXMatrixLookAtLH matView, Orig, Orient, Vector3(0, 1, 0)
        Case CM_LOOKATFRAME ' use position but look at target frame
        Case CM_WELDTOFRAME ' follow target frame
    End Select
    ' set view tranformation matrix
    D3Ddevice.SetTransform D3DTS_VIEW, matView
    ' set perspective transformation matrix
    D3DXMatrixPerspectiveFovLH matProjection, FOV, Aspect, NearClip, FarClip
    D3Ddevice.SetTransform D3DTS_PROJECTION, matProjection
End Sub


