VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cFrame"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False

Option Explicit

' a user friendly name for the frame
Public Tag As String

' the locations for this frame
Dim Loc() As New cLocation

' the mesh object
Public Mesh As D3DXMesh
' the mesh subsets
Dim Subset() As New cMeshSubset

' the collision test targets
Dim Collision() As cFrame

' the radius variable is used for sphere collision detection
Public Radius As Single

' the parent frame of a frame is a pointer to the frame this frame
' is derived from. works just like family trees - like you're derived
' from your mum so that's your parent. although in this case, it's only
' possible to have one parent (in real life, it's possible to have 2)
' this is recursive, so to find your grandparent you say Frame.Parent.Parent
Public ParentFrame As cFrame

' the frame contains an array of children, so each frame can have as
' many children as they want. this is recursive, so to find the
' grandchild of the frame you say Frame.ChildFrame(x).ChildFrame(y)
Private Child() As New cFrame

' string to store an error report
Private strReport As String



' initialize arrays - in this class, all elements with an index of 0
' aren't used, but they do exist
Private Sub Class_Initialize()
On Error Resume Next
    ReDim Loc(0)
    Set Loc(0).Parent = Me
    ReDim Subset(0)
    Subset(0).SubsetID = 0
    Set Subset(0).Parent = Me
    ReDim Collision(0)
End Sub



' get/set locations
Property Get Location(Index As Long) As cLocation
    Set Location = Loc(Index - 1)
End Property
Property Set Location(Index As Long, newLocation As cLocation)
    Set Loc(Index - 1) = newLocation
End Property



' get/let number of locations (origins & orienatations)
Property Get NumLocations() As Long
On Error Resume Next
    NumLocations = UBound(Loc) + 1
End Property
Property Let NumLocations(Num As Long)
Dim oldNum As Long
Dim l As Long
On Error Resume Next
    oldNum = UBound(Loc) + 1
    ReDim Preserve Loc(0 To Num - 1)
    If oldNum <= Num Then
        For l = oldNum To Num - 1
            Set Loc(l).Parent = Me
        Next
    End If
End Property



' adds a new location
Sub AddLocation(Origin As D3DVECTOR, Orientation As D3DVECTOR, Size As D3DVECTOR, Velocity As D3DVECTOR, Spin As D3DVECTOR, Growth As D3DVECTOR)
On Error Resume Next
Dim l As Long
    l = NumLocations + 1
    NumLocations = l
    With Loc(l - 1)
        Set .Parent = Me
        .Origin = Origin
        .Orientation = Orientation
        .Size = Size
        .Velocity = Velocity
        .Spin = Spin
        .Growth = Growth
    End With
End Sub



' this function transforms the frame using the origin, orientation
' velocity, rotation and time elapsed
' you must feed it the amount of time for the move in seconds
Sub RefreshLocations(timeElapsed As Single)
Dim l As Long
Dim vec As D3DVECTOR
On Error Resume Next
    For l = 0 To UBound(Loc)
        Loc(l).Refresh timeElapsed
    Next
End Sub



' the number of children belonging to the frame
Property Get NumChildren() As Long
On Error Resume Next
    NumChildren = UBound(Child)
End Property
Property Let NumChildren(Num As Long)
Dim oldNum As Long
Dim l As Long
On Error Resume Next
    oldNum = UBound(Child)
    ReDim Preserve Child(0 To Num)
    ' new children must have their parent told to them!
    If oldNum < Num Then
        For l = oldNum + 1 To Num
            Set Child(l).ParentFrame = Me
        Next
    End If
End Property



' child frame data
Property Get ChildFrame(Index As Long) As cFrame
On Error Resume Next
    Set ChildFrame = Child(Index)
End Property
Property Set ChildFrame(Index As Long, ChildFrame As cFrame)
On Error Resume Next
    Set Child(Index) = ChildFrame
End Property



' collision test data
Property Get CollisionTest(Index As Long) As cFrame
On Error Resume Next
    Set CollisionTest = Collision(Index)
End Property
Property Set CollisionTest(Index As Long, newCollisionTest As cFrame)
On Error Resume Next
    Set Collision(Index) = newCollisionTest
End Property



' get/let number of collision tests
Property Get NumCollisionTests() As Long
On Error Resume Next
    NumCollisionTests = UBound(Collision) + 1
End Property
Property Let NumCollisionTests(Num As Long)
On Error Resume Next
    ReDim Preserve Collision(0 To Num - 1)
End Property



' adds a new collision test
Sub AddCollisionTest(newCollisionTest As cFrame)
On Error Resume Next
    ReDim Preserve Collision(0 To UBound(Collision) + 1)
    Set Collision(UBound(Collision)) = newCollisionTest
End Sub


' adds a new subset
Private Sub AddSubset(Material As cMaterial, Texture As cTexture, SubsetID As Long)
On Error Resume Next
    ReDim Preserve Subset(0 To UBound(Subset) + 1)
    With Subset(UBound(Subset))
        Set .Parent = Me
        Set .Material = Material
        Set .Texture = Texture
        .SubsetID = SubsetID
    End With
End Sub



' adds to error report
Sub AddToReport(str As String)
On Error Resume Next
    strReport = strReport & str & vbNewLine
End Sub



' load the mesh from a .x file
Function LoadFromX(Filename As String, strReport As String) As X_ERROR
Dim MaterialBuffer As D3DXBuffer
Dim TextureName As String
Dim NumSubsets As Long
Dim tmpMaterial As D3DMATERIAL8
Dim l As Long
On Error GoTo LoadFailed
    ' set name of frame
    Tag = Filename
    ' create full filename
    Filename = XfilePath & Filename
    ' start report
    strReport = Now & " - Opening " & Filename & "." & vbNewLine
    ' load mesh
    Set Mesh = D3DX.LoadMeshFromX(Filename, D3DXMESH_MANAGED, D3Ddevice, Nothing, MaterialBuffer, NumSubsets)
    ' check for error
    If Mesh Is Nothing Then
        AddToReport "FATAL ERROR: " & Filename & " Not Found!"
        LoadFromX = XERR_FILENOTFOUND
        GoTo LoadFailed
    End If
    ' loop though each subset
    For l = 0 To NumSubsets - 1
        ' get material
        D3DX.BufferGetMaterial MaterialBuffer, l, tmpMaterial
        tmpMaterial.Ambient = tmpMaterial.Diffuse
        ' get texture
        TextureName = D3DX.BufferGetTextureName(MaterialBuffer, l)
        ' load material and texture
        ReDim Preserve Material(0 To UBound(Material) + 1)
        Material(UBound(Material)).Material = tmpMaterial
        Material(UBound(Material)).Tag = "Material_" & l
        If TextureName = "" Then
            AddSubset Material(UBound(Material)), Nothing, l
        Else
            ReDim Preserve Texture(0 To UBound(Texture) + 1)
            Texture(UBound(Texture)).Tag = TextureName
            If Texture(UBound(Texture)).LoadFromFile = False Then
                LoadFromX = XERR_SEEREPORT
                AddToReport "ERROR: Could not load texture " & Texture(UBound(Texture)).Tag
            End If
            AddSubset Material(UBound(Material)), Texture(UBound(Texture)), l
        End If
    Next
    Set MaterialBuffer = Nothing
    ' calculate radius
    D3DX.ComputeBoundingSphereFromMesh Mesh, Vector3(0, 0, 0), Radius
    If LoadFromX <> XERR_SEEREPORT Then LoadFromX = X_OK
    Exit Function
LoadFailed:
On Error Resume Next
    Set MaterialBuffer = Nothing
    If LoadFromX <> XERR_FILENOTFOUND Then LoadFromX = XERR_UNKNOWN
End Function



' render the mesh onto the device backbuffer
Sub Render()
Dim l As Long
On Error Resume Next
    For l = 1 To UBound(Subset)
        ' draw subset of mesh
        Subset(l).Render
    Next
End Sub



' cleans up the mess that this class makes
Private Sub Class_Terminate()
Dim i As Integer
On Error Resume Next
    ' unload objects
    ReDim Subset(0)
    Set Subset(0) = Nothing
    ReDim Location(0)
    Set Location(0) = Nothing
    ReDim Collision(0)
    Set Collision(0) = Nothing
    ' unload children recursively
    For i = LBound(Child) To UBound(Child)
        Set Child(i) = Nothing
    Next
End Sub
