Attribute VB_Name = "modEXFfile"

Option Explicit

' for deleting error logs
Private Declare Function DeleteFile Lib "kernel32" (ByVal lpFileName As String) As Long

' the .exf file version this module reads
Public Const EXF_FILE_VERSION = 1

' the dir to save files in
Public EXFfilePath As String

' possible errors when reading/writing .exf files
Public Enum EXF_FILE_ERR
    EXF_OK
    EXFERR_SEEREPORT
    EXFERR_FILENOTFOUND
    EXFERR_UNKNOWNFORMAT
    EXFERR_UNKNOWNVERSION
    EXFERR_UNKNOWN
End Enum

' strings containing extra file information
Public FileInfo As String
' error log
Private strReport As String

' the file number being used
Private n As Long






' get error log string
Property Get ErrorLog() As String
    ErrorLog = strReport
End Property




' this function loads a .exf file into the engine
Function LoadEXFfile(Filename As String, Optional Frame As cFrame = Nothing) As EXF_FILE_ERR
Dim curFrame As cFrame
Dim strIn As String
Dim tmpResult As EXF_FILE_ERR
    If EXFfilePath = "" Then EXFfilePath = App.Path & "\"
    Filename = EXFfilePath & Filename

On Error GoTo FileNotFound
    strReport = Now & " : Opening .exf file " & Filename & vbNewLine
    n = FreeFile
    Open Filename For Input As #n
    AddToReport "File opened Successfully, Reading File Header Info"
On Error GoTo Unknown

    tmpResult = ReadFileInfo
    If tmpResult <> EXF_OK Then
        LoadEXFfile = tmpResult
        GoTo General
    End If

On Error Resume Next
    
    If Frame Is Nothing Then
        Set curFrame = WorldFrame
    Else
        Set curFrame = Frame
    End If
    'Frame.NumChildren = 0
    
    AddToReport "Reading Geometry"
    Do
    
        Input #n, strIn
        Select Case UCase(strIn)
            Case "<CAMERA>"
                If ReadCameras = False Then
                    LoadEXFfile = EXFERR_SEEREPORT
                    AddToReport "Could not load camera " & Camera(UBound(Camera)).Tag & "."
                End If
            Case "<LIGHT>"
                If ReadLights = False Then
                    LoadEXFfile = EXFERR_SEEREPORT
                    AddToReport "Could not load light " & Light(UBound(Light)).Tag & "."
                End If
            Case "<SOUND>"
                If ReadSounds = False Then
                    LoadEXFfile = EXFERR_SEEREPORT
                    AddToReport "Could not load sound " & Sound(UBound(Sound)).Tag & "."
                End If
            Case "<FRAME>"
                ' add a child frame
                curFrame.NumChildren = curFrame.NumChildren + 1
                ' step down one level to the new child frame
                Set curFrame = curFrame.ChildFrame(curFrame.NumChildren)
                ' read the frame
                If ReadFrame(curFrame) = False Then
                    LoadEXFfile = EXFERR_SEEREPORT
                    AddToReport "Could not load frame " & curFrame.Tag & "."
                End If
            Case "</FRAME>"
                ' step up one level to the current frame's parent
                Set curFrame = curFrame.ParentFrame
            Case "<LOCATIONS>"
                If ReadLocations(curFrame) = False Then
                    LoadEXFfile = EXFERR_SEEREPORT
                    AddToReport "Could not load locations in frame " & curFrame.Tag & "."
                End If
'            Case "<COLLISION>"
'                If Readcollisions(curFrame) = False Then
'                    LoadEXFfile = EXFERR_SEEREPORT
'                    AddToReport "Could not load collision detection data in frame " & curFrame.Tag & "."
'                End If
            Case "<3DSOUND>"
                If Read3DSounds = False Then
                    LoadEXFfile = EXFERR_SEEREPORT
                    AddToReport "Could not load 3D sound in frame " & curFrame.Tag & "."
                End If
            Case "<COMMENT>"
                Do
                    Input #n, strIn
                Loop Until strIn = "</COMMENT>"
            Case "<END>"
                Exit Do
        End Select
    Loop Until EOF(1)
    AddToReport "Geometry loaded"

On Error Resume Next
    Close #n
    AddToReport "File load complete, file closed"
    If LoadEXFfile <> EXFERR_SEEREPORT Then LoadEXFfile = EXF_OK
    Exit Function
    
FileNotFound:
On Error Resume Next
    Close #n
    AddToReport "FATAL ERROR! File not found! Aborting file load!"
    LoadEXFfile = EXFERR_FILENOTFOUND
    Exit Function
UnknownFormat:
On Error Resume Next
    Close #n
    AddToReport "FATAL ERROR! File is of unknown format! Aborting file load!"
    LoadEXFfile = EXFERR_UNKNOWNFORMAT
    Exit Function
Unknown:
On Error Resume Next
    Close #n
    AddToReport "FATAL ERROR! An unknown error has occured during file load! Aborting file load"
    LoadEXFfile = EXFERR_UNKNOWN
    Exit Function
General:
On Error Resume Next
    AddToReport "ERROR! Aborting file load!"
    Close #n
End Function



' adds a line to the report
Private Sub AddToReport(strLine As String)
On Error Resume Next
    strReport = strReport & strLine & vbNewLine
End Sub



' reads the file header
Private Function ReadFileInfo() As EXF_FILE_ERR
Dim strIn As String
Dim strHeader As String
Dim Version As Byte
On Error GoTo ReadFailed
    Input #n, strHeader
    ' check version number
    If strHeader = "<ENGINEX>" Then
        Input #n, strIn
        Version = Val(strIn)
        AddToReport "File is a .exf file version " & Version
        If Version > EXF_FILE_VERSION Then
            ReadFileInfo = EXFERR_UNKNOWNFORMAT
            AddToReport "ERROR! This program can only open files up to .exf version " & EXF_FILE_VERSION
            Exit Function
        End If
        AddToReport "Version OK, now reading additional file information"
    Else
        AddToReport "ERROR! This file is of an unknown format!"
        ReadFileInfo = EXFERR_UNKNOWNFORMAT
        Exit Function
    End If
    FileInfo = ""
    ' read each line of additional information
    Do
        Input #n, strIn
        FileInfo = FileInfo & strIn & vbNewLine
    Loop Until strIn = "</ENGINEX>"
On Error Resume Next
    AddToReport "File Header Info Read OK"
    ReadFileInfo = EXF_OK
    Exit Function
ReadFailed:
On Error Resume Next
    AddToReport "ERROR! An unknown error occured while reading file header info!"
    ReadFileInfo = EXFERR_UNKNOWN
End Function



' reads a list of cameras
Private Function ReadCameras() As Boolean
Dim Tag As String
Dim Origin As D3DVECTOR, Orientation As D3DVECTOR
Dim FOV As Single, NearClip As Single, FarClip As Single, Aspect As Single
On Error GoTo ReadFailed
    Do
        ReDim Preserve Camera(0 To UBound(Camera) + 1)
        With Camera(UBound(Camera))
            Input #n, Tag
            If Tag = "</CAMERA>" Then Exit Do
            Input #n, .Origin.x, Origin.y, Origin.z, Orientation.x, Orientation.y, Orientation.z, FOV, NearClip, FarClip, Aspect
            .Tag = Tag
            .Origin = Origin
            .Orientation = Orientation
            .FOV = FOV
            .NearClip = NearClip
            .FarClip = FarClip
            .Aspect = Aspect
        End With
    Loop
    ReDim Preserve Camera(0 To UBound(Camera) - 1)
    ReadCameras = Not Err.Number
    Exit Function
ReadFailed:
    AddToReport "ERROR! Invalid camera definition!"
End Function



' reads a list of lights
Private Function ReadLights() As Boolean
Dim Tag As String
Dim Enabled As Boolean
Dim tmpLight As D3DLIGHT8
On Error GoTo ReadFailed
    Do
        ReDim Preserve Light(0 To UBound(Light) + 1)
        With tmpLight
            Input #n, Tag
            If Tag = "</LIGHT>" Then Exit Do
            Input #n, Enabled, .Type, .Ambient.a, .Ambient.r, .Ambient.g, .Ambient.b, .Diffuse.a, .Diffuse.r, .Diffuse.g, .Diffuse.b, .Specular.a, .Specular.r, .Specular.g, .Specular.b, .Position.x, .Position.y, .Position.z, .Direction.x, .Direction.y, .Direction.z, .Range, .Attenuation0, .Attenuation1, .Attenuation2, .Falloff, .theta, .phi
        End With
        With Light(UBound(Light))
            .Tag = Tag
            .Light = tmpLight
            .Index = UBound(Light)
            .Load
            .Enabled = Enabled
        End With
    Loop
    ReDim Preserve Light(0 To UBound(Light) - 1)
    ReadLights = Not Err.Number
    Exit Function
ReadFailed:
    AddToReport "ERROR! Invalid light definition!"
End Function



' reads a list of sounds
Private Function ReadSounds() As Boolean
Dim Tag As String
Dim Volume As Long
On Error GoTo ReadFailed
    Do
        Input #n, Tag
        If Tag = "</SOUND>" Then Exit Do
        Input #n, Volume
'        Sound(UBound(Sound)).Tag = Tag
'        If Sound(UBound(Sound)).LoadFromFile() = False Then GoTo ReadFailed
        If modAudio.AddSound(Tag) = False Then GoTo ReadFailed
        If Int(Val(Volume)) <> -1 Then Sound(UBound(Sound)).Volume = Volume
    Loop
    ReadSounds = Not Err.Number
    Exit Function
ReadFailed:
    AddToReport "ERROR! Invalid sound definition!"
End Function



' reads a list of 3d sounds
Private Function Read3DSounds() As Boolean
Dim BaseIndex As String, MinDist As Single, MaxDist As Single, OutsideVol As Long
On Error GoTo ReadFailed
    Do
        Input #n, BaseIndex
        If BaseIndex = "</3DSOUND>" Then Exit Do
        Input #n, MinDist, MaxDist, OutsideVol
        If modAudio.Add3Dsound(Sound(Int(Val(BaseIndex))), MinDist, MaxDist, OutsideVol) = False Then GoTo ReadFailed
    Loop
    Read3DSounds = Not Err.Number
    Exit Function
ReadFailed:
    AddToReport "ERROR! Invalid sound definition!"
End Function



' reads a frame definition
Private Function ReadFrame(Frame As cFrame) As Boolean
Dim Tag As String
Dim strRep As String
On Error GoTo ReadFailed
    With Frame
        Input #n, Tag
        .Tag = Tag
        If Tag <> "" Then If .LoadFromX(Tag, strRep) <> X_OK Then GoTo ReadFailed
    End With
    ReadFrame = Not Err.Number
    Exit Function
ReadFailed:
    AddToReport strRep
    AddToReport "ERROR! Invalid frame definition!"
End Function



' read locations into the specified frame
Private Function ReadLocations(Frame As cFrame) As Boolean
Dim str1 As String
Dim Origin As D3DVECTOR, Orientation As D3DVECTOR, Size As D3DVECTOR
Dim Velocity As D3DVECTOR, Spin As D3DVECTOR, Growth As D3DVECTOR
Dim IsFirstLocation As Boolean
On Error GoTo ReadFailed
    Frame.NumLocations = 1
    IsFirstLocation = True
    Do
        Input #n, str1
        If str1 = "</LOCATIONS>" Then Exit Do
        Input #n, Origin.y, Origin.z, Orientation.x, Orientation.y, Orientation.z, Size.x, Size.y, Size.z, Velocity.x, Velocity.y, Velocity.z, Spin.x, Spin.y, Spin.z, Growth.x, Growth.y, Growth.z
        Origin.x = Val(str1)
        If IsFirstLocation Then
            Frame.Location(1).SetVectors Origin, Orientation, Size, Velocity, Spin, Growth
            IsFirstLocation = False
        Else
            Frame.AddLocation Origin, Orientation, Size, Velocity, Spin, Growth
        End If
    Loop
    ReadLocations = Not Err.Number
    Exit Function
ReadFailed:
End Function












' this function saves a file in the si3d engine to disk
Function SaveEXFfile(Filename As String, Optional Frame As cFrame = Nothing) As EXF_FILE_ERR
On Error Resume Next
    If EXFfilePath = "" Then EXFfilePath = App.Path & "\"
    Filename = EXFfilePath & Filename
    DeleteFile Filename
On Error GoTo FileNotFound
    strReport = Now & " : Saving .exf file " & Filename & vbNewLine
    n = FreeFile
    Open Filename For Output As #n
    AddToReport "New File Created Successfully"
  
On Error GoTo Unknown
    AddToReport "Saving File Info Header"
    If WriteFileInfo = False Then SaveEXFfile = EXFERR_SEEREPORT
    AddToReport "Saving Cameras"
    If WriteCameras = False Then SaveEXFfile = EXFERR_SEEREPORT
    AddToReport "Saving Lights"
    If WriteLights = False Then SaveEXFfile = EXFERR_SEEREPORT
    AddToReport "Saving Sounds"
    If WriteSounds = False Then SaveEXFfile = EXFERR_SEEREPORT
    AddToReport "Saving 3D Sounds"
    If Write3Dsounds = False Then SaveEXFfile = EXFERR_SEEREPORT
    AddToReport "Saving Geometry"
    If Frame Is Nothing Then Set Frame = WorldFrame
    If WriteFrame(Frame, True) = False Then SaveEXFfile = EXFERR_SEEREPORT
    Write #n, "<END>"
    AddToReport "Closing File : " & Filename
    Close #n
    SaveEXFfile = EXF_OK
    Exit Function
    
On Error Resume Next
    Close #n
    AddToReport "File load complete, file closed"
    If SaveEXFfile <> EXFERR_SEEREPORT Then SaveEXFfile = EXF_OK
    Exit Function
    
FileNotFound:
On Error Resume Next
    Close #n
    AddToReport "FATAL ERROR! File not found! Aborting file load!"
    SaveEXFfile = EXFERR_FILENOTFOUND
    Exit Function
Unknown:
On Error Resume Next
    Close #n
    AddToReport "FATAL ERROR! An unknown error has occured during file load! Aborting file load"
    SaveEXFfile = EXFERR_UNKNOWN
    Exit Function
General:
On Error Resume Next
    AddToReport "ERROR! Aborting file load!"
    Close #n
End Function



' saves a file header
Private Function WriteFileInfo() As Boolean
Dim l As Long
On Error GoTo WriteFailed
    Write #n, "<ENGINEX>"
    Write #n, EXF_FILE_VERSION
    If FileInfo <> "" Then
        Write #n, FileInfo
    End If
    Write #n, "</ENGINEX>"
    WriteFileInfo = Not Err.Number
    Exit Function
WriteFailed:
    AddToReport "ERROR! Could not save file info header!"
End Function





' saves the cameras in the engine
Private Function WriteCameras() As Boolean
Dim l As Long
On Error GoTo WriteFailed
    If UBound(Camera) Then
        Write #n, "<CAMERA>"
        For l = 1 To UBound(Camera)
            With Camera(l)
                Write #n, .Tag, .Origin.x, .Origin.y, .Origin.z, .Orientation.x, .Orientation.y, .Orientation.z, .FOV, .NearClip, .FarClip, .Aspect
            End With
        Next
        Write #n, "</CAMERA>"
    End If
    WriteCameras = Not Err.Number
    Exit Function
WriteFailed:
    AddToReport "ERROR! Could not save camera definitions!"
End Function



' saves the lights in the engine
Private Function WriteLights() As Boolean
Dim l As Long
On Error GoTo WriteFailed
    If UBound(Light) Then
        Write #n, "<LIGHT>"
        For l = 1 To UBound(Light)
            With Light(l).Light
                Write #n, Light(l).Tag, Light(l).Enabled, .Type, .Ambient.a, .Ambient.r, .Ambient.g, .Ambient.b, .Diffuse.a, .Diffuse.r, .Diffuse.g, .Diffuse.b, .Specular.a, .Specular.r, .Specular.g, .Specular.b, .Position.x, .Position.y, .Position.z, .Direction.x, .Direction.y, .Direction.z, .Range, .Attenuation0, .Attenuation1, .Attenuation2, .Falloff, .theta, .phi
            End With
        Next
        Write #n, "</LIGHT>"
    End If
    WriteLights = Not Err.Number
    Exit Function
WriteFailed:
    AddToReport "ERROR! Could not save light definitions!"
End Function



' saves sounds in the engine
Private Function WriteSounds() As Boolean
Dim l As Long
On Error GoTo WriteFailed
    If UBound(Sound) Then
        Write #n, "<SOUND>"
        For l = 1 To UBound(Sound)
            If TypeOf Sound(l) Is cSound Then Write #n, Sound(l).Tag, -1  'Sound(l).Volume
        Next
        Write #n, "</SOUND>"
    End If
    WriteSounds = Not Err.Number
    Exit Function
WriteFailed:
    AddToReport "ERROR! Could not save sound definitions!"
End Function



' saves 3d sounds in the engine
Private Function Write3Dsounds() As Boolean
Dim l As Long
Dim m As Long
On Error GoTo WriteFailed
    If UBound(Sound) Then
        Write #n, "<3DSOUND>"
        For l = 1 To UBound(Sound)
            If TypeOf Sound(l) Is c3Dsound Then
                For m = 1 To UBound(Sound)
                    If TypeOf Sound(m) Is cSound Then
                        If Sound(l).BaseSound Is Sound(m) Then
                            With Sound(l)
                                Write #n, m, .MinDistance, .MaxDistance, .ConeOutsideVolume
                            End With
                        End If
                    End If
                Next
            End If
        Next
        Write #n, "</3DSOUND>"
    End If
    Write3Dsounds = Not Err.Number
    Exit Function
WriteFailed:
    AddToReport "ERROR! Could not save sound definitions!"
End Function



' saves the frames in the engine recursively
Private Function WriteFrame(Frame As cFrame, Optional FirstTime As Boolean = False) As Boolean
Dim l As Long
On Error GoTo WriteFailed
    WriteFrame = True
    With Frame
        ' write frame definition, skip if firsttime flag is true
        If Not FirstTime Then
            Write #n, "<FRAME>"
            Write #n, .Tag
            ' write locations
            If .NumLocations Then
                Write #n, "<LOCATIONS>"
                For l = 1 To .NumLocations
                    With .Location(l)
                    Write #n, .Origin.x, .Origin.y, .Origin.z, .Orientation.x, .Orientation.y, .Orientation.z, .Size.x, .Size.y, .Size.z, .Velocity.x, .Velocity.y, .Velocity.z, .Spin.x, .Spin.y, .Spin.z, .Growth.x, .Growth.y, .Growth.z
                    End With
                Next
                Write #n, "</LOCATIONS>"
            End If
        End If
        ' write children (recursive)
        If .NumChildren Then
            For l = 1 To .NumChildren
                WriteFrame = WriteFrame(.ChildFrame(l)) ' this carries errors up the tree
            Next
        End If
        ' end frame
        If Not FirstTime Then Write #n, "</FRAME>"
    End With
    If WriteFrame Then WriteFrame = Not Err.Number
    Exit Function
WriteFailed:
    WriteFrame = False
    AddToReport "ERROR! Could not save frame definition of " & Frame.Tag & "!"
End Function





