Attribute VB_Name = "modEnumeration"
Option Explicit

Type ENGINEX_DEVICE
    Hardware As Boolean
    Fullscreen As Boolean
    DisplayMode As D3DDISPLAYMODE
    Caps As D3DCAPS8
End Type

Enum HARDWARE_TEST_RESULT
    HTR_OK
    HTR_ERROR
    HTR_INVALIDFORMAT
    HTR_NOHARDWARE
End Enum

Private BestDevice As ENGINEX_DEVICE



' finds the best device possible
Function FindBestDevice(BestDevice As ENGINEX_DEVICE, Optional Fullscreen As Boolean = True) As Boolean
Dim Formats() As CONST_D3DFORMAT
Const NUM_FORMATS = 4
Dim ValidDevice() As ENGINEX_DEVICE
Dim tmpDevice As ENGINEX_DEVICE
Dim tmpD3D As Direct3D8
Dim SomeHardware As Boolean
Dim BestFormat As Long
Dim l As Long
On Error Resume Next
    ReDim ValidDevice(0)
    ' get a pointer to direct3d8
    Set tmpD3D = DX.Direct3DCreate
    If tmpD3D Is Nothing Then Exit Function
    ' create a list of possible formats, in order of preference
    ReDim Formats(1 To NUM_FORMATS)
    Formats(1) = D3DFMT_R5G6B5
    Formats(2) = D3DFMT_X1R5G5B5
    Formats(3) = D3DFMT_R8G8B8
    Formats(4) = D3DFMT_X8R8G8B8
    For l = 1 To NUM_FORMATS
        If D3D.CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, Format(l), Format(l), Not Fullscreen) = D3D_OK Then
            With tmpDevice
                .Hardware = False
                .DisplayMode.Format = l
                .Fullscreen = Fullscreen
            End With
            If D3D.CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, Format(l), Format(l), Not Fullscreen) = D3D_OK Then
                tmpDevice.Hardware = True
            End If
            ReDim Preserve ValidDevice(0 To UBound(ValidDevice) + 1)
            ValidDevice(UBound(ValidDevice)) = tmpDevice
        End If
    Next
    If UBound(ValidDevice) Then
        ' if some devices have hardware acceleration, kick out all the ones that don't
        For l = 1 To UBound(ValidDevice)
            If ValidDevice(l).Hardware Then SomeHardware = True
        Next
        If SomeHardware Then
            For l = 1 To UBound(ValidDevice)
                If Not ValidDevice(l).Hardware Then
                    ValidDevice(l) = ValidDevice(UBound(ValidDevice))
                    ReDim Preserve ValidDevice(0 To UBound(ValidDevice) - 1)
                    l = l - 1
                End If
            Next
        End If
        ' now choose the device with the best format
        BestFormat = NUM_FORMATS + 1
        For l = 1 To UBound(ValidDevice)
            If ValidDevice(l).DisplayMode.Format < BestFormat Then
                BestDevice = ValidDevice(l)
                BestFormat = BestDevice.DisplayMode.Format
            End If
        Next
        ' now get more information about the chosen best device
        With BestDevice
            .DisplayMode.Format = Formats(.DisplayMode.Format)
            .Fullscreen = Fullscreen
            If .Hardware Then
                tmpD3D.GetDeviceCaps D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, .Caps
            Else
                tmpD3D.GetDeviceCaps D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, .Caps
            End If
        End With
        FindBestDevice = True
    Else
        GoTo DeviceNotFound
    End If
    Set tmpD3D = Nothing
    Exit Function
DeviceNotFound:
On Error Resume Next
    FindBestDevice = False
    Set tmpD3D = Nothing
End Function



' checks if a display mode is valid
Function IsDisplayModeOK(testMode As D3DDISPLAYMODE) As Boolean
Dim NumModes As Long
Dim l As Long
Dim tmpMode As D3DDISPLAYMODE
Dim tmpD3D As Direct3D8
On Error Resume Next
    Set tmpD3D = DX.Direct3DCreate
    If tmpD3D Is Nothing Then Exit Function
    ' get number of modes
    NumModes = tmpD3D.GetAdapterModeCount(D3DADAPTER_DEFAULT)
    ' get each display mode
    For l = 0 To NumModes - 1
        tmpD3D.EnumAdapterModes D3DADAPTER_DEFAULT, l, tmpMode
        ' if there's a match then this is valid
        If tmpMode.Width = testMode.Width Then
            IsDisplayModeOK = True
            Exit For
        End If
    Next
Set tmpD3D = Nothing
End Function



' gets valid display modes
'Function GetValidDisplayModes(Optional Format As CONST_D3DFORMAT = 0) As D3DDISPLAYMODE()
'Dim NumModes As Long
'Dim l As Long
'Dim tmpMode As D3DDISPLAYMODE
'On Error Resume Next
'Dim tmpD3D As Direct3D8
'    ReDim GetValidDisplayModes(0)
'    Set tmpD3D = DX.Direct3DCreate
'    ' get number of modes
'    NumModes = D3D.GetAdapterModeCount(D3DADAPTER_DEFAULT)
'    ' get each display mode
'    For l = 0 To NumModes - 1
'        D3D.EnumAdapterModes D3DADAPTER_DEFAULT, l, tmpMode
'        ' get rid of unwanted formats
'        If Format Then If tmpMode.Format <> Format Then GoTo Next1
'        ' get rid of extra refresh rates
'        If tmpMode.RefreshRate Then GoTo Next1
'        ' add the display mode to the list
'        ReDim Preserve GetValidDisplayModes(0 To UBound(GetValidDisplayModes) + 1)
'        GetValidDisplayModes(UBound(GetValidDisplayModes)) = tmpMode
'Next1:
'    Next
'Set tmpD3D = Nothing
'End Function



' quick check for hardware acceleration, the result can be yes, no, or maybe, but in another format, or just error
Function CheckForHardwareAcceleration(BackbufferFormat As CONST_D3DFORMAT, Optional DisplayFormat As CONST_D3DFORMAT = 0, Optional Fullscreen As Boolean = True) As HARDWARE_TEST_RESULT
Dim tmpD3D As Direct3D8
On Error GoTo CheckFailed
    ' fill in the optional parameter
    If DisplayFormat = 0 Then DisplayFormat = BackbufferFormat
    ' get a pointer to a direct3d8 object
    Set D3D = DX.Direct3DCreate()
    If D3D Is Nothing Then GoTo CheckFailed
    ' check device for hardware acceleration
    Select Case D3D.CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, DisplayFormat, BackbufferFormat, Not Fullscreen)
        Case D3D_OK ' yes :)
            CheckForHardwareAcceleration = HTR_OK
        Case D3DERR_INVALIDDEVICE ' no :(
            CheckForHardwareAcceleration = HTR_NOHARDWARE
        Case D3DERR_NOTAVAILABLE ' maybe, try again...
            CheckForHardwareAcceleration = HTR_INVALIDFORMAT
        Case Else ' doh!
            CheckForHardwareAcceleration = HTR_ERROR
    End Select
    Set tmpD3D = Nothing
    Exit Function
CheckFailed:
On Error Resume Next
    CheckForHardwareAcceleration = HTR_ERROR
    Set tmpD3D = Nothing
End Function

